module 'data.path' {
   /** delimiter(Delim).
    * Succeeds if <code>Delim</code> is the atom of the path delimiter.
    * Always returns / for compatibility reason (/ is preferred on all platforms
    * by OS built-ins).
    * If you want the platform specific delimiter, use sys.os:path_delimiter. 
    */

   delimiter = '/'.

   /** separator(Sep).
    *  Succeeds if <code>Sep</code> is the atom of the path separator in PATH
    *  environment variables.
    *  Typically, : on Unixes and ; on Windows.
    */

   separator = sys.os:path_separator.

   /** concat(A, B, R).
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to paths.
    *  </prerequisite>
    *  If <code>B</code> is absolute, succeeds if <code>R</code> is <code>B</code>.
    *  If <code>B</code> is relative, concats paths <code>A</code> and <code>B</code>,
    *  separated with a path delimiter if needed.
    */

   concat(A, B, R) :-
      delimiter(Delim),
      (
         (
            is_absolute(B)
         ;
            A = ''
         )
      ->
         R = B
      ;
         data.atom:sub(A, _, 1, 0, Delim)
      ->
         data.atom:concat(A, B, R)
      ;
         data.atom:(
            concat(A, Delim, A0),
            concat(A0, B, R)
         )
      ).

   /** join(Paths, Path).
    *  <prerequisite>
    *  <code>Paths</code> is a list of instantiated paths.
    *  </prerequisite>
    *  Succeeds if <code>Path</code> is the concatenation of the elements of
    *  <code>Paths</code>, separated with path delimiters.
    */

   join(Paths, Path) :-
      delimiter(Delim),
      data.atom:join(Paths, Delim, Path).

   /** split(Path, Base, Filename).
    *  <prerequisite>
    *  <code>Path</code> is an instantiated path.
    *  </prerequisite>
    *  Succeeds if <code>Base</code> is the path of the directory of
    *  <code>Path</code> and <code>Filename</code> is the filename
    *  pointed by <code>Path</code> relative to this directory.
    */

   split(Path, Base, Filename) :-
      (
         data.atom:check(Base),
         data.atom:check(Filename)
      ->
         concat(Base, Filename, Path)
      ;
         delimiter(Delim),
         (
            data.atom:split_at_delim_right(Path, Delim, Base, Filename)
         ->
            true
         ;
            Base = '',
            Filename = Path
         )
      ).

   /** base(Path, Base).
    *  Equivalent to <code>split(Path, Base, _)</code>.
    */

   base(Path, Base) :-
      split(Path, Base, _).

   /** filename(Path, Filename).
    *  Equivalent to <code>split(Path, _, Filename)</code>.
    */

   filename(Path, Filename) :-
      split(Path, _, Filename).

   /** split_suffix(Path, Base, Suffix).
    *  <prerequisite>
    *  <code>Path</code> is an instantiated path.
    *  </prerequisite>
    *  Succeeds if <code>Base</code> is <code>Path</code> with
    *  the suffix removed, and <code>Suffix</code> is this suffix,
    *  dot excepted.
    */

   split_suffix(Path, Base, Suffix) :-
      delimiter(Delim),
      module Aux [Delim, Path, Base, Suffix] {
         split(Index) :-
            'data.number':(Index >= 0),
            'data.atom':sub(Path, Index, 1, _, C),
            (
               C = '.'
            ->
               'data.atom':sub(Path, 0, Index, _, Base),
               'data.atom':sub(Path, 'data.number':succ(Index), _, 0, Suffix)
            ;
               C \= Delim,
               this:split('data.number':pred(Index))
            ). 
      },
      (
         Aux:split('data.number':pred('data.atom':length(Path)))
      ->
         true
      ;
         Base = Path,
         Suffix = ''
      ).

   /** suffix(Path, Suffix).
    *  Equivalent to <code>split_suffix(Path, _, Suffix)</code>.
    */

   suffix(Path, Suffix) :-
      'data.path':split_suffix(Path, _, Suffix).

   /** chop_suffix(Path, Base).
    *  Equivalent to <code>split_suffix(Path, Base, _)</code>.
    */

   chop_suffix(Path, Base) :-
      'data.path':split_suffix(Path, Base, _).

   /** is_absolute(Path).
    *  Succeeds of <code>Path</code> is an absolute path.
    */

   is_absolute(Path) :-
      delimiter(Delim),
      'data.atom':sub(Path, 0, 1, _, Delim).


   trailing_path_delimiter(Without, With) :-
      (
         data.atom:check(Without),
         \+ data.atom:concat(_, '/', Without)
      ->
         data.atom:concat(Without, '/', With)
      ;
         data.atom:check(With),
         data.atom:concat(Without, '/', With)
      ->
         true
      ;
         With = Without
      ).
/*
   decompose(Filename, Path, Base, Suffix) :-
      (
         data.atom:check(Filename)
      ->
         (
            data.atom:split_at_delim_right(Filename, '/', Path, Basename)
         ->
            true
         ;
            Path = '',
            Basename = Filename
         ),
         (
            data.atom:split_at_delim_right(Basename, '.', Base, Suffix)
         ->
            true
         ;
            Base = Basename,
            Suffix = ''
         )
      ;
         data.atom:check(Path),
         data.atom:check(Base),
         data.atom:check(Suffix)
      ->
         (
            Suffix = ''
         ->
            Basename = Base
         ;
            Basename = data.atom:concat(data.atom:concat(Base, '.'), Suffix)
         ),
         (
            Path = ''
         ->
            Filename = Basename
         ;
            trailing_path_delimiter(Path, Fullpath),
            data.atom:concat(Fullpath, Basename, Filename)
         )
      ;
         throw('instantiation_error')
      ).

   change_suffix(Filename, New_suffix, New_filename) :-
      decompose(Filename, Path, Base, _Old_suffix),
      decompose(New_filename, Path, Base, New_suffix).
*/
}
