module emop.lexer [Impl] {
   positionable_stream(Stream, Lexems) :-
      Impl:lex(Stream, Lexems).

   utf8_stream(Stream, Filename, Lexems) :-
      Backtrackable = io.stream:backtrackable(Stream),
      Positionable = io.stream:text_position_utf8(Backtrackable, Filename),
      Lexems = positionable_stream(Positionable).
},

module Impl [Impl] {
   constructor: no/0,end_of_file/0,end_of_file/1,yes/0,end_of_line/1,line/0,comment/4,doc/0,long/0,unescaped/0,colon/3,file/2,error/2,atom/4,variable/0,number/3,escaped/0,string/3,close/3,functor/4,'*'/2.

   lex(Stream, Lexems) :-
      module Lexems [Stream, Impl] {
         read(Lexem) :-
            Stream:position(Pb),
            (
               Stream:read(C)
            ->
               true
            ),
            (
               C = end_of_file
            ->
               Lexem = end_of_file(Pb)
            ;
               Impl:blank(C)
            ->
               read(Lexem)
            ;
               C = '\n'
            ->
               Lexem = end_of_line(Pb)
            ;
               C = '%'
            ->
               read_line_comment(Comment),
               Stream:position(Pe),
               Lexem = comment(line, Comment, Pb, Pe)
            ;
               C = '/',
               Stream:read('*')
            ->
               (
                  Stream:read('*')
               ->
                  read_long_comment(Comment),
                  Stream:position(Pe),
                  Lexem = comment(doc, Comment, Pb, Pe)
               ;
                  read_long_comment(Comment),
                  Stream:position(Pe),
                  Lexem = comment(long, Comment, Pb, Pe)
               )
            ;
               (
                  data.term:(
                     C >= 'A',
                     C =< 'Z'
                  )
               ;
                  data.term:(
                     C >= 'a',
                     C =< 'z'
                  )
               ;
                  C = '_'
               ;
                  data.number:(data.char:code(C) >= 0x80)
               )
            ->
               read_ident_tail(C, Atom),
               read_atom_close(Atom, unescaped, Pb, Lexem)
            ;
               C = '.',
               Stream:read('(')
            ->
               Stream:position(Pe),
               Lexem = '.('(Pb, Pe)
            ;
               C = '.',
               Stream:read(C0),
               data.char:decimal_digit(C0, _)
            ->
               read_tail_numeric(Tail),
               read_decimal_end(['0'], [C0 | Tail], Pb, Value),
               Stream:position(Pe),
               Lexem = number(Value, Pb, Pe)
            ;
               (
                  C = ','
               ;
                  C = ';'
               )
            ->
               Stream:position(Pe),
               Lexem = atom(C, unescaped, Pb, Pe)
            ;
               (
                  Impl:symbol_char(C)
               ;
                  C = '.'
               )
            ->
               read_symbol_tail(C, Atom),
               Stream:position(Pe),
               Lexem = atom(Atom, unescaped, Pb, Pe)
            ;
               C = '0',
               Stream:read('.'),
               Stream:read(C0),
               data.char:decimal_digit(C0, _)
            ->
               read_tail_numeric(Tail),
               read_decimal_end(['0'], [C0 | Tail], Pb, Value),
               Stream:position(Pe),
               Lexem = number(Value, Pb, Pe)
            ;
               C = '0'
            ->
               (
                  Stream:read(C0),
                  (
                     (
                        C0 = 'x'
                     ;
                        C0 = 'X'
                     )
                  ->
                     read_hexa(Value)
                  ;
                     (
                        C0 = 'b'
                     ;
                        C0 = 'B'
                     )
                  ->
                     read_binary(Value)
                  ;
                     C0 = '\''
                  ->
                     Stream:read(C1),
                     (
                        C1 = '\\'
                     ->
                        data.char:read_escape(Stream, ValueC)
                     ;
                        ValueC = C1
                     ),
                     Value = data.char:code(ValueC)
                  ;
                     (
                        (
                           C0 = 'o'
                        ;
                           C0 = 'O'
                        )
                     ->
                        Stream:read(C1)
                     ;
                        data.char:decimal_digit(C0, _),
                        C1 = C0
                     )
                  ->
                     read_octal(C1, Value)
                  )
               ->
                  true
               ;
                  Value = 0
               ),
               Stream:position(Pe),
               Lexem = number(Value, Pb, Pe)
            ;
               data.char:decimal_digit(C, _)
            ->
               read_decimal(C, Pb, Value),
               Stream:position(Pe),
               Lexem = number(Value, Pb, Pe)
            ;
               C = '\''
            ->
               String = read_quoted_string(Pb, _) {
                  do('\'').
               },
               data.string:atom(String, Atom),
               read_atom_close(Atom, escaped, Pb, Lexem)
            ;
               C = '\"'
            ->
               String = read_quoted_string(Pb, _) {
                  do('\"')
               },
               Stream:position(Pe),
               Lexem = string(String, Pb, Pe)
            ;
               C = '`'
            ->
               String = read_quoted_string(Pb, Delim) {
                  do('`').
                  do('\'').
               },
               data.string:atom(String, Atom),
               Lexem = Delim match (
                  '`'
               ->
                  atom(Atom, variable, Pb, Pe)
               ;
                  _
               ->
                  read_atom_close(Atom, unescaped, Pb)
               )
            ;
               (
                  C = '('
               ;
                  C = '['
               ;
                  C = '{'
               )
            ->
               Stream:position(Pe),
               data.term:decomposition(Lexem, [C, Pb, Pe])
            ;
               (
                  C = ')'
               ;
                  C = ']'
               ;
                  C = '}'
               )
            ->
               Stream:position(Pe),
               Lexem = close(C, Pb, Pe)
            ;
               Stream:position(Pe),
               throw(error('Unknown char', file(Pb, Pe)))
            ).

         read_line_comment(Comment) :-
            module Accu [Stream] {
               get(C) :-
                  (
                     Stream:read(C)
                  ->
                     C \= '\n'
                  ).
            },
            data.list:accumulate(Accu, String),
            data.string:atom(String, Comment).

         read_long_comment(Comment) :-
            module [Stream, Pb] {
               skip_beginning_of_line :-
                  (
                     Stream:read match (
                        (' '; '\t')
                     ->
                        !,
                        skip_beginning_of_line
                     ;
                        '*'
                     ->
                        \+ Stream:read('/'),
                        !
                     ;
                        _
                     ->
                        !
                     )
                  ;
                     true
                  ).
               do(List) :-
                  (
                     Stream:read(C0)
                  ->
                     (
                        C0 = 'end_of_file'
                     ->
                        Stream:position(Pe),
                        throw(
                           error('Unterminated comment', file(Pb, Pe))
                        )
                     ;
                        C0 = '\n'
                     ->
                        skip_beginning_of_line,
                        do(List)
                     ;
                        C0 = '*',
                        Stream:read('/')
                     ->
                        List = []
                     ;
                        List = [C0 | T],
                        do(T)
                     )
                  ).
            }:do(String),
            data.string:atom(String, Comment).

         read_atom_close(Atom, Escaped, Pb, Lexem) :-
            Stream:position(Pe),
            (
               Stream:read('(')
            ->
               Lexem = functor(Atom, Escaped, Pb, Pe)
            ;
               Lexem = atom(Atom, Escaped, Pb, Pe)
            ).

         sequence_of_dots =
            data.list:accumulate(module [Stream] { get('.') :- Stream:read('.') }).

         dotted_symbol(C, M, Atom) :-
            module Accu_ident [Stream, M] {
               get(L) :-
                  (
                     Stream:read(C),
                     C \= 'end_of_file',
                     (
                        M:do(C)
                     ->
                        L = [C]
                     ;
                        C = '.',
                        (
                           T = sequence_of_dots,
                           T \= []
                        ;
                           Stream:read(C0),
                           M:do(C0),
                           T = [C0]
                        )
                     ->
                        L = [C | T]
                     )
                  ->
                     true
                  ;
                     fail
                  ).
            },
            data.list:accumulate_sublists(Accu_ident, T),
            data.string:atom(data.list:append(C, T), Atom).

         read_ident_tail(C) = dotted_symbol([C]) [] {
            do(C) :-
               data.term:(
                  C \= 'end_of_file',
                  (
                     C >= 'a',
                     C =< 'z'
                  ;
                     C >= 'A',
                     C =< 'Z'
                  ;
                     C >= '0',
                     C =< '9'
                  ;
                     C = '_'
                  ;
                     C = '\''
                  ;
                     data.number:(data.char:code(C) >= 0x80)
                  )
               ).
         }.

         read_symbol_tail(C, Atom) :-
            (
               C = '.'
            ->
               Firsts = [
                  C
               |
                  sequence_of_dots
               ]
            ;
               Firsts = [C]
            ),
            Atom = dotted_symbol(Firsts) [Impl] {
               do(C) :-
                  Impl:symbol_char(C).
            }.

         read_hexa(Value) :-
            module Aux [Stream] {
               read(V, R) :-
                  (
                     data.char:hexadecimal_digit(Stream:read, V0)
                  ->
                     this:read('data.number':(V * 0x10 + V0), R)
                  ;
                     V = R
                  ).
            },
            Aux:read(0, Value).

         read_binary(Value) :-
            module Aux [Stream] {
               read(V, R) :-
                  (
                     data.char:bit(Stream:read, V0)
                  ->
                     this:read('data.number':(V * 0b10 + V0), R)
                  ;
                     V = R
                  ).
            },
            Aux:read(0, Value).

         read_octal(C, Value) :-
            module Aux [Stream] {
               read(V, R) :-
                  (
                     data.char:octal_digit(Stream:read, V0)
                  ->
                     this:read('data.number':(V * 0o10 + V0), R)
                  ;
                     V = R
                  ).
            },
            (
               data.char:octal_digit(C, V0)
            ->
               Aux:read(V0, Value)
            ;
               Value = 0
            ).

         read_decimal(C, Pb, Value) :-
            read_tail_numeric(Tail_int_part),
            Int_part = [C | Tail_int_part],
            (
               Stream:read('.'),
               read_numeric(Float_part)
            ->
               true
            ;
               true
            ),
            read_decimal_end(Int_part, Float_part, Pb, Value).

          read_decimal_end(Int_part, Float_part, Pb, Value) :-
            (
               Stream:read('e')
            ->
               (
                  Float_part = ['0']
               ->
                  true
               ;
                  true
               ),
               (
                  Stream:read('-')
               ->
                  (
                     read_numeric(Tail_exp_part)
                  ->
                     Exp_part = ['-' | Tail_exp_part]
                  ;
                     Stream:position(Pe),
                     throw(
                        error('Ill-formed number', file(Pb, Pe))
                     )
                  )
               ;
                  read_numeric(Exp_part)
               ->
                  true
               ;
                  Stream:position(Pe),
                  throw(
                     error('Ill-formed number', file(Pb, Pe))
                  )
               )
            ;
               Exp_part = ['0']
            ),
            (
               Float_part = []
            ->
               data.string:number(Int_part, Value)
            ;
               data.list:append(Int_part, ['.' | Float_part], Partial),
               data.list:append(Partial, ['e' | Exp_part], String),
               data.string:number(String, Value)
            ).

         read_numeric(String) :-
            Stream:read(C0),
            data.char:decimal_digit(C0, _),
            String = [C0 | Tail],
            read_tail_numeric(Tail).

         read_tail_numeric(String) :-
            module Accu_digit [Stream] {
               get(C) :-
                  (
                     Stream:read(C),
                     data.char:decimal_digit(C, _)
                  ->
                     true
                  ).
            },
            data.list:accumulate(Accu_digit, String).
            
         read_quoted_string(Pb0, Delim, Delimiter, String) :-
            module Accu_string [Stream, Delimiter, Accu_string, Pb0] {
               get(C) :-
                  (
                     Stream:position(Pb),
                     Stream:read(C0),
                     (
                        C0 = end_of_file
                     ->
                        throw(error('Unterminated string', file(Pb0, Pb)))
                     ;
                        Delimiter:do(C0)
                     ->
                        Stream:read(C0),
                        C = C0
                     ;
                        C0 = '\\'
                     ->
                        (
                           (
                              true
                           ;
                              Stream:read('\r')
                           ),
                           Stream:read('\n') 
                        ->
                           Accu_string:skip_space(C)
                        ;
                           data.char:read_escape(Stream, C)
                        ->
                           true
                        ;
                           Stream:position(Pe),
                           throw(
                              error('Ill-formed escape character', file(Pb, Pe))
                           )
                        )
                     ;
                        C = C0
                     )
                  ->
                     true
                  ).

               skip_space(C) :-
                  (
                     Stream:read(C0),
                     (
                        C0 = ' '
                     ;
                        C0 = '\t'
                     )
                  ->
                     Accu_string:skip_space(C)
                  ;
                     Accu_string:get(C)
                  ).
            },
            data.list:accumulate(Accu_string, String),
            (
               Stream:read(Delim),
               Delimiter:do(Delim)
            ->
               true
            ;
               Stream:position(Pe),
               throw(error('Unterminated string', file(Pb0, Pe)))
            ).

      }.

   blank(C) :-
      (
         C = ' '
      ;
         C = '\t'
      ;
         C = '\r'
      ),
      !.

   symbol_char(C) :-
      (
         (
            C = ':'
         ;
            C = '+'
         ;
            C = '-'
         ;
            C = '*'
         ;
            C = '/'
         ;
            C = '\\'
         ;
            C = '='
         ;
            C = '<'
         ;
            C = '>'
         ;
            C = '|'
         ;
            C = '@'
         ;
            C = '#'
         ;
            C = '?'
         ;
            C = '!'
         ;
            C = '&'
         ;
            C = '~'
         ;
            C = '$'
         ;
            C = '^'
         )
      ->
         true
      ;
         fail
      ).
}
